/**
 * @ingroup   file68_devel
 * @file      SC68file.c
 * @author    Ben(jamin) Gerard <ben@sashipa.com>
 * @date      1999/05/11
 * @brief     SC68 general purpose OS-DEPENDANT function
 * @version   $Id: SC68os_general.c,v 1.1.1.1 2001/12/24 20:50:10 ben Exp $
 */

/*
 *                         sc68 - Disk file access
 *         Copyright (C) 2001 Ben(jamin) Gerard <ben@sashipa.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>

#if defined(WIN32) && !defined(CYGNUS)
# include <direct.h>
#else
# include <unistd.h>
#endif


#include "file68/SC68os_general.h"
#include "file68/SC68error.h"
#include "file68/SC68utils.h"


#ifndef FL_PATH_MAX
# define FL_PATH_MAX 1024
#endif

/*  Debug output ( printf-like format )
 */

#ifdef WIN32
# include <windows.h>
#endif

void SC68os_pdebug(char *format, ...)
{
#ifdef _DEBUG
#ifdef WIN32
  char buffer[1024];
#endif
  va_list list;
  va_start(list, format);
#ifdef WIN32
  _vsnprintf(buffer, sizeof(buffer), format, list);
  buffer[sizeof(buffer) - 1] = 0;
  OutputDebugString(buffer);
#else
  vfprintf(stderr, format, list);
#endif
  va_end(list);
#endif
}

/****************************************************************************
* SC68 RESOURCE OPERATIONS
****************************************************************************/

#ifdef WIN32
static char window_sc68_path[FL_PATH_MAX];
#endif

/* Construct OS compatible filename for a ressource.
*/
char *SC68os_resource_name(SC68rsc_t rsc, char *fname)
{
  static char s[1024];
  char *home;

#ifdef WIN32
  home = window_sc68_path;
#else
  home = getenv("HOME");
#endif

  if (!home) {
    SC68error_add("Cannot find home path");
    return 0;
  }
  s[sizeof(s) - 1] = 0;
  if (s != home)
    strncpy(s, home, sizeof(s));

#ifdef WIN32
  SC68utils_strcat(s, "/sc68/", sizeof(s));
#else
  SC68utils_strcat(s, "/.sc68/", sizeof(s));
#endif

  switch (rsc) {
    /* For 68000 external replay */
  case SC68rsc_replay:
    SC68utils_strcat(s, "Replay/", sizeof(s));
    SC68utils_strcat(s, fname, sizeof(s));
    break;

    /* Browse info */
  case SC68rsc_browseinfo:
    SC68utils_strcat(s, "Browse/", sizeof(s));
    SC68utils_strcat(s, fname, sizeof(s));
    SC68utils_strcat(s, ".txt", sizeof(s));
    break;

    /* HTML generate for browse info */
  case SC68rsc_html:
    SC68utils_strcat(s, "html/", sizeof(s));
    SC68utils_strcat(s, fname, sizeof(s));
    SC68utils_strcat(s, ".html", sizeof(s));
    break;

    /* Config File */
  case SC68rsc_config:
    SC68utils_strcat(s, "Config/", sizeof(s));
    SC68utils_strcat(s, fname, sizeof(s));
    SC68utils_strcat(s, ".txt", sizeof(s));
    break;

    /* Unix specific resource file */
  case SC68rsc_unix:
    SC68utils_strcat(s, "Unix/", sizeof(s));
    SC68utils_strcat(s, fname, sizeof(s));
    break;

    /* sc68 sample resource file */
  case SC68rsc_sample:
    SC68utils_strcat(s, "Sample/", sizeof(s));
    SC68utils_strcat(s, fname, sizeof(s));
    break;

    /* sc68 dynamic library (for winamp gl about) */
  case SC68rsc_dll:
    SC68utils_strcat(s, "Dll/", sizeof(s));
    SC68utils_strcat(s, fname, sizeof(s));
    break;

  default:
    SC68error_add("SC68os_resource_name : Unknown type:%d", (int) rsc);
    return NULL;
  }
  if (s[sizeof(s) - 1]) {
    SC68error_add("SC68os_resource_name : file name too long");
    return NULL;
  }
  return s;
}

/*  Construct OS compatible filename for a resource
 *  and open it in given mode.
*/
FILE *SC68os_open_resource(SC68rsc_t rsc, char *fname, char *openmode)
{
  FILE *f;
  char *s = SC68os_resource_name(rsc, fname);

  if (s == NULL || (f = SC68_fopen(s, openmode), f == NULL)) {
    SC68error_add("SC68os_open_resource(%s,%s) : Failed ", s, openmode);
    return NULL;
  }

/* DO SOMETHING LIKE MIMESET OR WHAT ? */
/*  if(*openmode=='w' && rsc==SC68rsc_html)
  {
  }
*/
  return f;
}




/****************************************************************************
* FILE-NAME OPERATION
****************************************************************************/

/*  Print a file name into file, with on the fly '.'<->'/' convertion
*/
int SC68os_fput_fname(FILE * f, char *fname)
{
  char tmp[512];
  strncpy(tmp, fname, sizeof(tmp));
  tmp[sizeof(tmp) - 1] = 0;
  fputs(SC68os_os2posix_fname(tmp), f);
  return 0;
}

/* Convert file name string from OS-style to POSIX compatible
*/
char *SC68os_os2posix_fname(char *filename)
{
#ifdef WIN32
  int i;
  i = 0;
  for (; filename[i]; i++)
    if (filename[i] == '\\')
      filename[i] = '/';
#endif
  return filename;
}

/* Convert file name string from POSIX to OS-style compatible
*/
char *SC68os_posix2os_fname(char *filename)
{
#ifdef WIN32
  int i;
  for (i = 0; filename[i]; i++)
    if (filename[i] == '/')
      filename[i] = '\\';
#endif
  return filename;
}

static char *delchar(char *name, int n)
{
  int len = strlen(name);
  if (n > len)
    n = len;
  memmove(name, name + n, len - n + 1);
  return name;
}

static char *remove_double_slash(char *name)
{
  int i;
  for (i = 0; name[i];) {
    if (name[i] == '/' && name[i + 1] == '/')
      delchar(name + i + 1, 1);
    else
      i++;
  }
  return name;
}

static char *remove_single_point(char *name)
{
  char *p, *s = name + 1;
  while (p = strchr(s, '.'), p != NULL) {
    if (p[1] == 0 || p[1] == '/') {
      if (p[-1] == '/')
	delchar(p, 1);
    }
    s = p + 1;
  }
  return name;
}

static char *remove_double_point(char *name)
{
  char *p, *s = name;
  while (p = strchr(s, '.'), p != NULL) {
    if (p[1] == '.' && p > name + 1 && p[-1] == '/'
	&& (p != name + 2 || name[2] != '.') && (p <= name + 2
						 || p[-3] != '.'
						 || p[-2] != '.')) {
      int n = 4;
      p--;
      while (p > name && p[-1] != '/') {
	p--;
	n++;
      }
      delchar(p, n);
      s = p;
    } else
      s = p + 1;
  }
  return name;
}


static char *canonikal(char *name)
{
  if (name == NULL)
    return NULL;
  remove_single_point(name);
  remove_double_slash(name);
  remove_double_point(name);

  return name;
}

/* Create absolute path name, with curdir + filename
*/
char *SC68os_absolute_fname(char *filename)
{
  static char name[FL_PATH_MAX];
  char *home = NULL;
  int maxlen = sizeof(name) - 1;
  name[0] = 0;

  if (filename == NULL)
    return NULL;
  filename = SC68os_os2posix_fname(filename);
/*  SC68os_pdebug("make_absolute(%s)\n",filename);*/

#ifdef WIN32

  /* Drive letter present */
  if (isalpha(filename[0]) && filename[1] == ':') {
    /* Already absolute */
    if (filename[2] == '/')
      strncpy(name, filename, maxlen);

    /* Relatif to a given drive */
    else {
      if (_getdcwd(toupper(name[0]) - 'A' + 1, name, maxlen))
	name[0] = 0;
      SC68utils_strcat(name, "/", maxlen);
      SC68utils_strcat(name, filename, maxlen);
    }
  }

  /* Network name  */
  else if (filename[0] == '/' && filename[1] == '/')
    strncpy(name, filename, maxlen);

  else {
    if (_getcwd(name, maxlen) == NULL)
      name[0] = 0;
    /* Absolute current drive */
    if (filename[0] == '/') {
      if (isalpha(name[0]) && name[1] == ':')
	strncpy(name + 2, filename, maxlen - 2);
      else
	strncpy(name, filename, maxlen);
    }

    /* Relatif current drive */
    else {
      if (isalpha(name[0]) && name[1] == ':') {
	SC68utils_strcat(name, "/", maxlen);
	SC68utils_strcat(name, filename, maxlen);
      } else
	strncpy(name, filename, maxlen);
    }
  }
  SC68os_os2posix_fname(name);
  if (name[0] == '/' && name[1] == '/')
    canonikal(strchr(name + 2, '/'));
  else
    canonikal(name);

#else

  if (filename[0] != '/') {
    home = getenv("PWD");
    if (home == NULL)
      home = getcwd(name, maxlen);
    else
      strncpy(name, home, maxlen);
  }
  if (home == NULL)
    return strncpy(name, filename, maxlen);
  SC68utils_strcat(name, "/", maxlen);
  SC68utils_strcat(name, canonikal(filename), maxlen);
  canonikal(name);

#endif

  name[maxlen] = 0;
  SC68os_posix2os_fname(name);
/*  SC68os_pdebug("make absolute path:\n<%s>\n<%s>\n",filename,name);*/
  return name;

}

/* Initialize executable path.
 * Typically path is argv[0] from int main(int argc, char **argv)
*/
int SC68os_get_exec_path(char *path)
{

#ifdef WIN32
  char *s;
  int maxlen = sizeof(window_sc68_path) - 1;
  if (path[0] == '"')
    strncpy(window_sc68_path, path + 1, maxlen);
  else
    strncpy(window_sc68_path, path, maxlen);
  SC68os_os2posix_fname(window_sc68_path);
  if (s = strrchr(window_sc68_path, '/'), s != NULL)
    *s = 0;
  else
    *window_sc68_path = 0;
#else
  char *home;
  home = getenv("HOME");
  if (home == NULL)
    return SC68error_add("Can't find environment variable HOME");
#endif

  return 0;
}
