/**
 * @ingroup   io68_devel
 * @file      paulaemul.c
 * @author    Ben(jamin) Gerard <ben@sashipa.com>
 * @date      1998/07/18
 * @brief     Paula emulator (Amiga soundchip)
 * @version   $Id: paulaemul.c,v 1.1.1.1 2001/12/24 20:50:10 ben Exp $
 */
 
/*
 *                 sc68 - Paula emulator (Amiga soundchip)
 *         Copyright (C) 1999 Ben(jamin) Gerard <ben@sashipa.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */ 

#define PAULA_CALCUL_TABLE 0

#if PAULA_CALCUL_TABLE
# include <math.h>
# include <stdio.h>
#endif

#include "io68/paulaemul.h"

#define REPLAY_FRQ_MAX 48000
#define REPLAY_FRQ_MIN 8000
#define DEFAULT_REPLAY_FRQ 44100
#define MAX_SPL_PER_PASS (REPLAY_FRQ_MAX/50)
#define LN_10_OVER_10 0.230258509299
#define PL_N_DECIBEL 65

#define CT_FIX       PAULA_CT_FIX
#define PL_VOL_FIX   16
#define PL_VOL_MUL   (1<<PL_VOL_FIX)
#define PL_MIX_FIX   (PL_VOL_FIX+1+8-16) /*(PL_VOL_FIX+2+8-16) */

#if PAULA_CALCUL_TABLE
static const s16 Tvol[] = {
  0, 1, 3, 4, 6, 7, 9, 10, 12, 13, 15, 16, 18, 20, 21, 23,
  25, 27, 29, 31, 33, 35, 37, 39, 41, 43, 45, 48, 50, 52, 55, 58,
  60, 63, 66, 69, 72, 75, 78, 82, 85, 89, 93, 97, 101, 105, 110,
  115, 120, 126, 132, 138, 145, 153, 161, 170, 181, 192, 206,
  221, 241, 266, 301, 361, 450
};
static u32 volume[65];
#else
static const u32 volume[65] = {
  0x00000,0x006ec,0x00c9e,0x011e8,0x016fe,0x01c15,0x020a0,0x02588,
  0x029e5,0x02ec4,0x0332b,0x0376e,0x03c0c,0x04067,0x04462,0x0489d,
  0x04d1b,0x0510f,0x05537,0x05995,0x05d3d,0x0610b,0x06501,0x06920,
  0x06d6b,0x070c0,0x0755a,0x078ed,0x07c9b,0x08067,0x08450,0x08857,
  0x08c7e,0x08f55,0x093b2,0x09832,0x09b45,0x09e68,0x0a33b,0x0a687,
  0x0a9e4,0x0ad53,0x0b0d3,0x0b466,0x0b80b,0x0bbc3,0x0bf8e,0x0c36c,
  0x0c75f,0x0cb66,0x0cf82,0x0d198,0x0d5d4,0x0da26,0x0dc57,0x0e0ca,
  0x0e30d,0x0e7a3,0x0e9f7,0x0eeb1,0x0f117,0x0f5f6,0x0f86f,0x0fd73,
  0x10000,
};
#endif

static u32 replay_frq, dividand;

u8 paula[256];
paulav_t paulav[4];
int paula_dmacon;               /* Internal DMA control register copy */
int paula_intena;               /* Internal interrupt enable register copy */
int paula_intreq;               /* Internal interrupt request register copy */

/******************************************************
*                  Set replay frequency               *
******************************************************/

u32 PL_set_replay_frq(u32 f)
{
  replay_frq = f < REPLAY_FRQ_MIN ? REPLAY_FRQ_MIN :
      (f > REPLAY_FRQ_MAX ? REPLAY_FRQ_MAX : f);
  dividand = (u32) ((PAULA_FRQ * (double) (1 << CT_FIX)) / replay_frq);
  return replay_frq;
}

#if PAULA_CALCUL_TABLE
static u32 calc_volume(u32 vol, u32 mult)
{
  double r;

  if (vol == 0)
    return 0;
  r = (double) mult / exp((double) Tvol[64 - vol] / 100.0);
  return (u32) r;
}

static void init_volume(void)
{
  int i;

  for (i = 0; i < 65; i++) {
    volume[i] = calc_volume(i, PL_VOL_MUL);
    fprintf(stderr, "XXX:%08x\n", volume[i]);
  }
}
#endif

/******************************************************
*                     PAULA reset                     *
******************************************************/

int PL_reset(void)
{
  int i;

  for (i = 0; i < sizeof(paula); i++)
    paula[i] = 0;
  for (i = 0; i < 4; i++) {
    paulav[i].adr = 0;
    paulav[i].start = 0;
    paulav[i].end = 2;
  }
  paula_dmacon = 1 << 9;    /* DMA general activate, DMA audio desactivate */
  paula_intreq = 0;         /* No interrupt request */
  paula_intena = 1 << 14;   /* Master interrupt enable, audio int disable */
  return 0;
}

/******************************************************
*                      PAULA init                     *
******************************************************/

int PL_init(void)
{
  PL_reset();
  PL_set_replay_frq(DEFAULT_REPLAY_FRQ);
#if PAULA_CALCUL_TABLE
  init_volume();
#endif
  return 0;
}

/* Mix with laudio channel data (1 char instead of 2) */
static void mix_zero(u32 * b, unsigned n, unsigned N, int shift)
{
  s32 o;
  u8 *p;
  paulav_t *w;
  u16 *b2;

  p = paula + PAULA_VOICE(N);
  w = paulav + N;

  /* Reload internal when interrupt is DENIED */
  if (
      (paula_intreq |
       ~((paula_intena << (8 * sizeof(int) - 1 - 14) >> (8 * sizeof(int) - 1))
         & paula_intena)) & (1 << (N + 7))) {
    w->adr = w->start =
        (((p[1] << 16) | (p[2] << 8) | p[3]) & 0x7FFFE) << PAULA_CT_FIX;
    w->end = ((p[4] << 8) | p[5]) << (1 + PAULA_CT_FIX);
    w->end = (w->end == 0) ? (1 << (17 + PAULA_CT_FIX)) : w->end;
    w->end += w->start;
  }
  paula_intreq |= 1 << (N + 7);

  /* nothing more to do */
  if (n == 0) {
    return;
  }

  b2 = (u16 *) b;
  if (shift) {
    ++b2;
  }

  o = p[9] & 127;
  if (o >= 64) {
    o = 64;
  }
  o = volume[o];
  o = (int) o * (int) (s8)p[0xA] /*^0x80 */ ;

/*  if (!o) {
    return;
  }*/
  o = 0;  //$$$


  o >>= PL_MIX_FIX;
  /*o <<= shift; */
  do {
  /**b++ += o;*/

    *b2 += o;
    b2 += 2;
  } while (--n);
}

static void mix_one(u32 * b, unsigned n, unsigned N, u8 * mem, int shift)
{
  paulav_t *w;
  u8 *p, last;
  unsigned stp, vol, per, readr, reend, end, hasloop;
  u32 adr;
  u16 *b2;

  /* nothing to do */
  if (n == 0)
    return;

  hasloop = 0;

  /* volume & period */
  p = paula + PAULA_VOICE(N);
  w = paulav + N;

  vol = p[9] & 127;
  if (vol >= 64)
    vol = 64;
  vol = volume[vol];

  per = ((p[6] << 8) + p[7]);
  if (per == 0)
    per = 1;
  stp = dividand / per;

// /* audio int enable for this voice */
// if(paula_intena & ((paula_intena&~paula_intreq)<<(14-7-N)) & (1<<14) )
// {
//  /* Audio interrupt enable for this voice : No internal reload */
//  readr = v->start;
//  reend = v->end;
// }
// else
  {
    /* Audio irq disable for this voice :
     * Internal will be reload at end of block
     */
    readr = (((p[1] << 16) | (p[2] << 8) | p[3]) & 0x7FFFE) << PAULA_CT_FIX;
    reend = ((p[4] << 8) | p[5]) << (1 + PAULA_CT_FIX);
    reend = (reend == 0) ? (1 << (17 + PAULA_CT_FIX)) : reend;
    reend += readr;
  }

  if (reend < readr) {
/*		printf("voice %d reend<readr !!!!!!!!!!!\n");*/
    return;
  }

  adr = w->adr;
  end = w->end;

/*	printf("V(%d) %03u:%03u -> %03u:%03u\n",N,adr>>PAULA_CT_FIX,end>>PAULA_CT_FIX,readr>>PAULA_CT_FIX,reend>>PAULA_CT_FIX);*/

  if (end < adr) {
/*		printf("V(%d) adr>end %03u<%03u!!!!!!!!!!!\n",N,adr>>PAULA_CT_FIX,end>>PAULA_CT_FIX);*/
    return;
  }

  b2 = (u16 *) b;
  if (shift) {
    ++b2;
  }

  /* mix stereo */
  do {
    int v, ov;
    int low;

    low = adr & ((1 << CT_FIX) - 1);
    if (adr >= end) {
      unsigned relen = reend - readr;

      hasloop = 1;
      adr = readr + adr - end;
      end = reend;
      while (adr >= end)
        adr -= relen;
    } {
      u32 i = adr >> CT_FIX;

      last = mem[i++];
      v = (int) (s8) last /*^ 0x80 */ ;
      if ((i << CT_FIX) >= end) {
        i = readr >> CT_FIX;
      }
      ov = (int) (s8) mem[i]; /*^ 0x80 */ ;
    }

    /* Linear interpol */
    v = (ov * low + v * ((1 << CT_FIX) - low)) >> CT_FIX;
    v *= vol;
    v >>= PL_MIX_FIX;
  /**b++ += v<<shift;*/

    /*
    {
      int vsave = v;
      int i, coef;

      for(i=0; i<15; i++) {
        w->v[i] = w->v[i+1];
      }
      w->v[i] = vsave;

      v = 0;
      coef = 0;
      for(i=0; i<16; i++) {
        v += w->v[i];
        coef += 1;
      }
      v /= coef;
    }
    */

    *b2 += v /*<<shift */ ;
    b2 += 2;

    adr += stp;
  } while (--n);

  if (adr >= end) {
    unsigned relen = reend - readr;

    hasloop = 1;
    adr = readr + adr - end;
    end = reend;
    while (adr >= end)
      adr -= relen;
  }

  p[0xA] = last + (last << 8);
  w->adr = adr;
  if (hasloop) {
    w->start = readr;
    w->end = end;
  }
}

/******************************************************
*                  PAULA process                      *
******************************************************/

void PL_mix(u32 * b, u8 * mem68, u8 * mem68end, u32 n)
{
  int i;

  if (n == 0)
    return;

  /* Clear buffer */
  {
    u32 *b2, n2;

    b2 = b;
    n2 = n;
    do {
      *b2++ = 0;
    } while (--n2);
  }

/*	printf("-------------\nMIX: %X\n",paula_dmacon&15);*/
  for (i = 0; i < 4; i++) {
    if ((paula_dmacon >> 9) & (paula_dmacon >> i) & 1) {
      mix_one(b, n, i, mem68, (i & 1) << 4);
    } else {
      mix_zero(b, n, i, (i & 1) << 4);
    }
  }
}
