/**
 * @ingroup   io68_devel
 * @file      ymemul.c
 * @author    Ben(jamin) Gerard <ben@sashipa.com>
 * @date      1998/06/24
 * @brief     YM-2149 emulator
 * @version   $Id: ymemul.c,v 1.1.1.1 2001/12/24 20:50:10 ben Exp $
 */

/*
 *                         sc68 - YM-2149 emulator
 *           Copyright (C) 1999 Ben(jamin) Gerard <ben@sashipa.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "config.h"
#include "io68/ymemul.h"

static u16 ymout[16*16*16] =
#include "io68/ymoutput.inc"
/*#include "io68/ymoutorg.h"*/

#define PLAYER_FRQ_MIN 48
#define REPLAY_FRQ_MAX 48000
#define REPLAY_FRQ_MIN 8000
#define DEFAULT_REPLAY_FRQ 44100
#define MAX_SPL_PER_PASS (REPLAY_FRQ_MAX/PLAYER_FRQ_MIN)
#define MAX_VOLUME_ENTRY 2048

static u32   mixbuf[MAX_SPL_PER_PASS], *mixptr;
static u32   envbuf[MAX_SPL_PER_PASS], *envptr;
static u32 noisebuf[MAX_SPL_PER_PASS], *noiseptr;

#ifdef _DEBUG
  static int noiseCnt;
  static int envCnt;
  static int mixCnt;
#endif

static u32 voiceCut = 0xFFF;

typedef struct
{
  u32 cycle;
  u8  reg,v;
} entry_t;

typedef struct
{
  u32 lc;                       /* Last cycle mixed */
  entry_t *c;                   /* Free entry */
  entry_t e[MAX_VOLUME_ENTRY];  /* Static register entry list */
} entrylist_t;

static entrylist_t envelopR, noiseR, mixerR;

static u32 env_model[8][32] =
{
  /* Cyclic Envelop */
  { 15,14,13,12,11,10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0,
    15,14,13,12,11,10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0 },
  { 15,14,13,12,11,10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0,
     0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15 },
  {  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,
     0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15 },
  {  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,
    15,14,13,12,11,10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0 },
  /* Non Cyclic Envelop */
  { 15,14,13,12,11,10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
  { 15,14,13,12,11,10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0,
    15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15 },
  {  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,
    15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15 },
  {  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
};

static struct
{
  int cyclic;
  u32 *model;
} env_model_def[16] =

{
  {0,&env_model[4][0]}, /* 0 \_ */
  {0,&env_model[4][0]}, /* 1 \_ */
  {0,&env_model[4][0]}, /* 2 \_ */
  {0,&env_model[4][0]}, /* 3 \_ */

  {0,&env_model[7][0]}, /* 4 /_ */
  {0,&env_model[7][0]}, /* 5 /_ */
  {0,&env_model[7][0]}, /* 6 /_ */
  {0,&env_model[7][0]}, /* 7 /_ */

  {1,&env_model[0][0]}, /* 8 \\ */
  {0,&env_model[4][0]}, /* 9 \_ */
  {1,&env_model[1][0]}, /* A \/ */
  {0,&env_model[5][0]}, /* B \- */
  {1,&env_model[2][0]}, /* C // */
  {0,&env_model[6][0]}, /* D /- */
  {1,&env_model[3][0]}, /* E /\ */
  {0,&env_model[7][0]}, /* F /_ */
};

/*static float replay_frq;*/
static unsigned ifrq;
static int cycle_per_spl, cycle_per_spl_fake;
static unsigned noisefake=0, mixerfake=0, envfake=0;

ym2149_t ym;

static void reset_entry(entrylist_t *l)
{
  l->lc = 0;
  l->c = l->e;
}

static void add_entry(entrylist_t *l, u32 reg, u32 v, u32 cycle)
{
  entry_t *p = l->e+MAX_VOLUME_ENTRY;
  if( l->c != p ) p=l->c++;
  else p--;
  p->cycle = cycle;
  p->reg = reg;
  p->v = v;
}

static u32 get_entry(entrylist_t *l, u32 reg, u32 cycle)
{
  entry_t *p;

  /* Found last minus cycle in write list */
  for(p=l->c-1; p>=l->e; p--)
    if( p->reg==reg ) return p->v;

  /* Not found : Get local */
  return ym.data[reg];
}

static void subcycle_entry(entrylist_t *l, u32 subcycle)
{
  entry_t *p;
  l->lc -= subcycle;
  for(p=l->e; p<l->c; p++) p->cycle -= subcycle;
}


/******************************************************
*                  Set replay frequency               *
******************************************************/

float YM_set_replay_frq(float f)
{
  float replay_frq;
  replay_frq = f<REPLAY_FRQ_MIN ? REPLAY_FRQ_MIN :
               (f>REPLAY_FRQ_MAX ? REPLAY_FRQ_MAX : f);
  ifrq = (unsigned)replay_frq;
  replay_frq = (float)ifrq;
  cycle_per_spl = (unsigned)8000000/ifrq;
  cycle_per_spl_fake = (unsigned)8000000%ifrq;
  return replay_frq;
}

/*
static u32 my_cast(double a)
{
  if(a>=4294967296.0) a = fmod(a,4294967296.0);
  return (u32)a;
}
*/

u32 Stp(u32 per)
{
  const u64 limit = (u64)1<<32;
  u64 a;
  if(!per) per=1;
  /* Or may be it is that, I can't remember.
  ++per;
  */
  a = (u64)0x1e8480aae6000 / (u64)(ifrq * per);
  if(a>=limit)
  {
    u32 b;
    b = (u32)a |  0x80000000;
    b &=         ~0x40000000;
    return b;
  }
  return (u32)a;
}

u32 Env_Stp(u32 per)
{
  u64 stp2;
  if(!per) per = 1; /* See above */
  /*  */
  stp2 = ((u64)125000<<27) / (u64)(ifrq*per);
  return (u32)stp2;
}

/******************************************************
*                  Yamaha reset                       *
******************************************************/

int YM_reset(void)
{
  int i;
  for(i=0; i<256; i++) ym.data[i]=0;
  ym.data[YM_MIXER] = 077;
  ym.data[YM_ENVTYPE] = 0x0a;
  ym.ctrl       = 0;
  ym.env_ct     = 0;
  ym.noise_gen  = 0x80000001;
  ym.noise_ct   = 0;
  ym.noise_stat = -1;
  ym.voice_ctA = 0;
  ym.voice_ctB = 0;
  ym.voice_ctC = 0;
  noisefake=mixerfake=envfake=0;
  /*voiceCut = 0xFFF;*/
  reset_entry(&mixerR);
  reset_entry(&noiseR);
  reset_entry(&envelopR);
  return 0;
}

/******************************************************
*                  Yamaha init                        *
******************************************************/

static void init_env( void)
{
  int i,j;
  for(i=0; i<8; i++) for(j=0; j<32; j++)
  {
    int v;
    v = env_model[i][j]&15;
    env_model[i][j] = v | (v<<4) | (v<<8);
  }
  return;
}

static void  init_output_level(int new_level)
{
  int i, mult = (new_level<<14)/(ymout[0xfff]+1);
  for(i=0; i<4096; i++) ymout[i] = (ymout[i]*mult)>>14;
  return;
}

int YM_init( void )
{
  init_env();
  init_output_level(0xD000);
  YM_reset();
  YM_set_replay_frq(DEFAULT_REPLAY_FRQ);
  return 0;
}

/******************************************************
*                  Noise generator                    *
******************************************************/

/*
 * Generate noise for ncycle ( 8Mhz )
 * - stp is step for this noise period
 * - msk is mixer mask for noise
 * - mix in noiseptr & update it
 *
 * -> return # cycle not mixed
 *
 */

static int generate_noise(int ncycle, u32 stp, u32 msk )
{
  u32 v;
  s32 noise_gen,ct,stat;
  u32 *b;

  /* Not enought cycle !!! */
  if(ncycle<cycle_per_spl) return ncycle;

  /* All inits */
  ct        = ym.noise_ct;
  noise_gen = ym.noise_gen;
  stat      = ym.noise_stat;
  b         = noiseptr;
  v         =  stat | msk;

  /* Bring the noise */
  do
  {
    ct += stp;
    if(ct<0)
    {
      /* --- Based on MAME :) --- */

      ct ^= 0x80000000; /* ??? */
    /* The Random Number Generator of the 8910 is a 17-bit shift */
    /* register. The input to the shift register is bit0 XOR bit2 */
    /* (bit0 is the output). */
    /* I invert shift from right to left a use msb in order to */
    /* use signed extension !!! */

      /* bit 17 (bit 14) = bit 0 ^ bit 2 ( bit 31 ^ bit 29 ) */
      noise_gen |= (unsigned)(noise_gen^(noise_gen<<2))>>17;
      noise_gen >>= 31-17;    /* Clear ORed waste bit*/
      noise_gen <<= 31-17+1;  /* Shift */
      stat = (signed)noise_gen>>31;
      v = stat | msk;
    }
    *b++ = v;

    ncycle    -= cycle_per_spl;
    noisefake += cycle_per_spl_fake;
    if(noisefake >= ifrq)
    {
       ncycle--;
       noisefake -= ifrq;
    }
  } while(ncycle>=cycle_per_spl);

#ifdef _DEBUG
  noiseCnt += b-noiseptr;
#endif

  /* Save value for next pass */
  noiseptr = b;
  ym.noise_gen = noise_gen;
  ym.noise_ct = ct;
  ym.noise_stat = stat;

  /* return not mixed cycle */
  return ncycle;
}

static void cycle_noise_gen( u32 cycle )
{
  entry_t *p;
  u32 lastcycle;
  unsigned per, msk, stp;
  noiseptr = noisebuf;

  per = ym.data[YM_NOISE];
  stp = Stp(per&0x3F);
  msk = ((ym.data[YM_MIXER]&010) ? 0xf00 : 0) |
        ((ym.data[YM_MIXER]&020) ? 0x0f0 : 0) |
        ((ym.data[YM_MIXER]&040) ? 0x00f : 0);

  for(p=noiseR.e, lastcycle=noiseR.lc; p<noiseR.c; p++)
  {
    lastcycle = p->cycle-generate_noise(p->cycle-lastcycle,stp,msk);
    if(p->reg==YM_NOISE)
    {
      per = p->v;
      stp = Stp(per&0x3F);
    }
    else
      msk = ((p->v&010) ? 0xf00 : 0) |
            ((p->v&020) ? 0x0f0 : 0) |
            ((p->v&040) ? 0x00f : 0);
  }

  ym.data[YM_NOISE] = per;
  lastcycle = cycle - generate_noise(cycle-lastcycle,stp,msk);

  noiseR.lc = lastcycle;
  noiseR.c = noiseR.e;
}

/******************************************************
*                Envelop generator                    *
******************************************************/

static int generate_env_cyclic(int ncycle, u32 *model, u32 stp)
{
  u32 *b;
  u32 ct;

  /* Not enought cycle !!! */
  if(ncycle<cycle_per_spl) return ncycle;

  b  = envptr;
  ct = ym.env_ct;
  do
  {
    *b++ = model[ct>>27];
    ct += stp;
    ncycle  -= cycle_per_spl;
    envfake += cycle_per_spl_fake;
    if(envfake >= ifrq)
    {
       ncycle--;
       envfake -= ifrq;
    }
  } while(ncycle>=cycle_per_spl);

  /* $$$  Must correct the env countr here */

#ifdef _DEBUG
  envCnt += b-envptr;
#endif

  /* Save value for next pass */
  envptr    = b;
  ym.env_ct = ct;

  return ncycle;
}

static int generate_env_noncyclic(int ncycle, u32 *model, u32 stp)
{
  u32 *b;
  u32 ct;

  /* Not enought cycle !!! */
  if(ncycle<cycle_per_spl) return ncycle;

  b  = envptr;
  ct = ym.env_ct;
  do
  {
    *b++ = model[ct>>27];
    ct = (ct+stp) | (ct&0x80000000) ;
    ncycle  -= cycle_per_spl;
    envfake += cycle_per_spl_fake;
    if(envfake >= ifrq)
    {
       ncycle--;
       envfake -= ifrq;
    }
  } while(ncycle>=cycle_per_spl);

  /* $$$  Must correct the env countr here */

#ifdef _DEBUG
  envCnt += b-envptr;
#endif

  /* Save value for next pass */
  envptr    = b;
  ym.env_ct = ct;

  return ncycle;
}

static void cycle_env_gen( u32 cycle )
{
  entry_t *p;
  u32 lastcycle;
  u32 stp,*model;
  int cyclic;

  envptr = envbuf;
  {
    int imodel;
    imodel = ym.data[YM_ENVTYPE]&0xF;
    model  = env_model_def[imodel].model;
    cyclic = env_model_def[imodel].cyclic;
  }
  stp    = Env_Stp(ym.data[YM_ENVL] | (ym.data[YM_ENVH]<<8));

  for(p=envelopR.e, lastcycle=envelopR.lc; p<envelopR.c; p++)
  {
    lastcycle = p->cycle -
      ( cyclic ? generate_env_cyclic(p->cycle-lastcycle,model,stp)
               : generate_env_noncyclic(p->cycle-lastcycle,model,stp));

    ym.data[p->reg] = p->v;
    if(p->reg==YM_ENVTYPE)
    {
      int imodel = p->v&0xF;
      model  = env_model_def[imodel].model;
      cyclic = env_model_def[imodel].cyclic;
      ym.env_ct = 0;
    }
    else
      stp    = Env_Stp(ym.data[YM_ENVL] | (ym.data[YM_ENVH]<<8));
  }

  lastcycle = cycle -
    ( cyclic ? generate_env_cyclic(cycle-lastcycle,model,stp)
             : generate_env_noncyclic(cycle-lastcycle,model,stp));

  envelopR.lc = lastcycle;
  envelopR.c = envelopR.e;
}

/******************************************************
*                Sound generator                      *
******************************************************/

static int generate_mixer( int ncycle, u32 smsk, u32 emsk, u32 vols )
{
  s32 ctA,  ctB,  ctC;
  u32 *o,*e,*b;

  /* Not enought cycle !!! */
  if(ncycle<cycle_per_spl) return ncycle;

  /* init buffer address */
  o=noiseptr;
  e=envptr;
  b=mixptr;

  /* Mixer steps & couters */
  ctA = ym.voice_ctA;
  ctB = ym.voice_ctB;
  ctC = ym.voice_ctC;

  do
  {
    u32 sq;
    sq  = (ctA>>31)&0xF00;
    sq += (ctB>>31)&0x0F0;
    sq += (ctC>>31)&0x00F;
    sq |= smsk;
    sq &= *o++;
    sq &= ((*e++)&emsk) | vols;
    sq &= voiceCut;

    *b++ = ymout[sq];
    ctA += ym.stpA;
    ctB += ym.stpB;
    ctC += ym.stpC;

    ncycle     -= cycle_per_spl;
    mixerfake += cycle_per_spl_fake;
    if(mixerfake >= ifrq)
    {
       ncycle--;
       mixerfake -= ifrq;
    }
  } while(ncycle>=cycle_per_spl);

  /* $$$ Add error correction here  */

#ifdef _DEBUG
  mixCnt += b-mixptr;
#endif

  /* Save value for next pass */
  ym.voice_ctA = ctA;
  ym.voice_ctB = ctB;
  ym.voice_ctC = ctC;
  noiseptr=o;
  envptr=e;
  mixptr=b;

  return ncycle;
}


static void cycle_mix_gen( u32 cycle )
{
  entry_t *p;
  u32 lastcycle,smsk,emsk,hfmsk,vols;
  int per;
  const int hflimit = 5; /* High frequencies generate parasite noise */

  mixptr   = mixbuf;
  envptr   = envbuf;
  noiseptr = noisebuf;
  hfmsk    = 0;

  per = ym.data[YM_PERL(0)] | ((ym.data[YM_PERH(0)]&0xF)<<8);
  if(per<hflimit) hfmsk |= 0xF00;
  else            hfmsk &= 0x0FF;
  ym.stpA = Stp(per);

  per = ym.data[YM_PERL(1)] | ((ym.data[YM_PERH(1)]&0xF)<<8);
  if(per<hflimit) hfmsk |= 0x0F0;
  else            hfmsk &= 0xF0F;
  ym.stpB = Stp(per);

  per = ym.data[YM_PERL(2)] | ((ym.data[YM_PERH(2)]&0xF)<<8);
  if(per<hflimit) hfmsk |= 0x00F;
  else            hfmsk &= 0xFF0;
  ym.stpC = Stp(per);

  /* 3 voice sound mask */
  smsk = ((ym.data[YM_MIXER]&001) ? 0xF00 : 0) |
         ((ym.data[YM_MIXER]&002) ? 0x0F0 : 0) |
         ((ym.data[YM_MIXER]&004) ? 0x00F : 0);


  /* 3 voices buzz or lvl mask */
  emsk = vols = 0;
  {
    int v;
    v = ym.data[YM_VOL(0)]&0x1F;
    if(v&0x10) emsk |= 0xF00;
    else       vols |= v<<8;

    v = ym.data[YM_VOL(1)]&0x1F;
    if(v&0x10) emsk |= 0x0F0;
    else       vols |= v<<4;

    v = ym.data[YM_VOL(2)]&0x1F;
    if(v&0x10) emsk |= 0x00F;
    else       vols |= v;
  }

  for(p=mixerR.e, lastcycle=mixerR.lc; p<mixerR.c; p++)
  {
    int v;
    lastcycle = p->cycle -
                generate_mixer(p->cycle-lastcycle, smsk|hfmsk, emsk, vols);
    v = ym.data[p->reg] = p->v;
    switch(p->reg)
    {
      case YM_VOL(0):
        v    &= 0x1F;
        vols &= 0x0FF;
        if(v&0x10)
          emsk |= 0xF00;
        else
        {
          emsk &= 0x0FF;
          vols |= v<<8;
        }
        break;

      case YM_VOL(1):
        v    &= 0x1F;
        vols &= 0xF0F;
        if(v&0x10)
          emsk |= 0x0F0;
        else
        {
          emsk &= 0xF0F;
          vols |= v<<4;
        }
        break;

      case YM_VOL(2):
        v    &= 0x1F;
        vols &= 0xFF0;
        if(v&0x10)
          emsk |= 0x00F;
        else
        {
          emsk &= 0xFF0;
          vols |= v;
        }
        break;

      case YM_MIXER:
        smsk = ((v&001) ? 0xf00 : 0) |
               ((v&002) ? 0x0f0 : 0) |
               ((v&004) ? 0x00f : 0);
        break;

      case YM_PERL(0): case YM_PERH(0):
        per =  ym.data[YM_PERL(0)] | ((ym.data[YM_PERH(0)]&0xF)<<8);
        if(per<hflimit) hfmsk |= 0xF00;
        else            hfmsk &= 0x0FF;
        ym.stpA = Stp(per);
        break;

      case YM_PERL(1): case YM_PERH(1):
        per =  ym.data[YM_PERL(1)] | ((ym.data[YM_PERH(1)]&0xF)<<8);
        if(per<hflimit) hfmsk |= 0x0F0;
        else            hfmsk &= 0xF0F;
        ym.stpB = Stp(per);
        break;

      case YM_PERL(2): case YM_PERH(2):
        per =  ym.data[YM_PERL(2)] | ((ym.data[YM_PERH(2)]&0xF)<<8);
        if(per<hflimit) hfmsk |= 0x00F;
        else            hfmsk &= 0xFF0;
        ym.stpC = Stp(per);
        break;
    }
  }
  lastcycle = cycle - generate_mixer(cycle-lastcycle, smsk|hfmsk, emsk, vols);
  mixerR.lc = lastcycle;
  mixerR.c = mixerR.e;
}

/******************************************************
*                  Yamaha process                     *
*               return # mixed sample                 *
******************************************************/
unsigned YM_mix(u32 cycle2mix)
{
  /* $$$ Must check buffer overflow here */

#ifdef _DEBUG
  /* $$$ Should not happen */
  mixCnt = noiseCnt = envCnt = 0;
#endif

  cycle_noise_gen(cycle2mix);
  cycle_env_gen(cycle2mix);
  cycle_mix_gen(cycle2mix);

#ifdef _DEBUG
  /* $$$ Should not happen */
  if (mixCnt != noiseCnt || mixCnt != envCnt) {
    BREAKPOINT68;
  }
#endif

  return ( (unsigned)mixptr - (unsigned)mixbuf) >> 2;
}

/*************************************
*         Write in YM register       *
*************************************/
void YM_writereg(u8 reg, u8 v, u32 cycle)
{
  switch(reg)
  {
    case YM_PERL(0): case YM_PERH(0):
    case YM_PERL(1): case YM_PERH(1):
    case YM_PERL(2): case YM_PERH(2):
    case YM_VOL(0): case YM_VOL(1): case YM_VOL(2):
      add_entry( &mixerR, reg, v,  cycle );
      break;

    case YM_ENVL: case YM_ENVH: case YM_ENVTYPE:
      add_entry( &envelopR, reg, v, cycle );
      break;

    case YM_MIXER: /* Reg 7 modify both noise gen */
                   /* & mixer */
      add_entry( &mixerR, reg, v, cycle );
    case YM_NOISE:
      add_entry( &noiseR, reg, v, cycle );
      break;

    default:
      break;
  }
}

/********************************************
*         Read from YM register      *
********************************************/
u8 YM_readreg(u8 reg, u32 cycle)
{
    /*return ym.data[reg];*/

  switch(reg)
  {
    case YM_PERL(0): case YM_PERH(0):
    case YM_PERL(1): case YM_PERH(1):
    case YM_PERL(2): case YM_PERH(2):
    case YM_VOL(0): case YM_VOL(1): case YM_VOL(2):
      return get_entry( &mixerR, reg, cycle );

    case YM_ENVL: case YM_ENVH: case YM_ENVTYPE:
      return get_entry( &envelopR, reg, cycle );

    case YM_NOISE:
    case YM_MIXER:
      /* Reg 7 is used in 2 lists */
      /* Better choice is noise   */
      return get_entry( &noiseR, reg, cycle );

    default:
      break;
  }
  return 0;
}

/****************************************** *
*         Change YM cycle count base        *
********************************************/
void YM_subcycle(u32 subcycle)
{
  subcycle_entry(&mixerR,   subcycle);
  subcycle_entry(&envelopR, subcycle);
  subcycle_entry(&noiseR,   subcycle);
}

/******************************************************
*                  Yamaha get buffer                  *
******************************************************/
u32 *YM_get_buffer( void )
{
  return mixbuf;
}

/******************************************************
*                 Yamaha get activated voices         *
******************************************************/
int YM_get_activevoices(void)
{
  return ((voiceCut>>8)&1) | ((voiceCut>>4)&2) | (voiceCut&4);
}

/******************************************************
*                 Yamaha set activated voices         *
******************************************************/
void YM_set_activeVoices(int v)
{
  int cut = 0;
  if (v&1) cut |= 0xF00;
  if (v&2) cut |= 0x0F0;
  if (v&4) cut |= 0x00F;
  voiceCut = cut;
}



